/*****************************************************************************
* luaEngine Lua Engine for Qt
* Copyright (C) 2018 Syping
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*****************************************************************************/

#define LUA_LIB
#include "LuaEngine.h"
#include <QTextStream>

LuaEngine::LuaEngine(QObject *parent, bool loadBaseLibraries) : QObject(parent)
{
    L = luaL_newstate();
    if (loadBaseLibraries)
        luaL_openlibs(L);
    pushFunction("luaEngineVersion", luaEngineVersion_p);
}

LuaEngine::~LuaEngine()
{
    lua_close(L);
}

lua_State* LuaEngine::luaState()
{
    return L;
}

void LuaEngine::loadBaseLibraries()
{
    luaL_openlibs(L);
}

int LuaEngine::luaEngineVersion_p(lua_State *L_p)
{
    pushVariant(L_p, "0.1");
    return 1;
}

bool LuaEngine::executeLuaScript(const QByteArray &data)
{
    int returnCode = luaL_dostring(L, data.data());
    return (returnCode == 0) ? true : false;
}

bool LuaEngine::executeLuaScript(QIODevice *device, bool closeDevice)
{
    QByteArray data;
    if (!device->isOpen()) {
        if (device->open(QIODevice::ReadOnly)) {
            data = device->readAll();
            if (closeDevice)
                device->close();
            return executeLuaScript(data);
        }
    }
    else {
        data = device->readAll();
        if (closeDevice)
            device->close();
        return executeLuaScript(data);
    }
    return false;
}

bool LuaEngine::executeLuaFunction(const char *name, bool requireReturn)
{
    return executeLuaFunction(L, name, requireReturn);
}

bool LuaEngine::executeLuaFunction(lua_State *L_p, const char *name, bool requireReturn)
{
    int returnCount = (requireReturn) ? 1 : 0;
    lua_getglobal(L_p, name);
    return (lua_pcall(L_p, 0, returnCount, 0) == 0) ? true : false;
}

bool LuaEngine::executeLuaFunction(const char *name, const QVariant &argument, bool requireReturn)
{
    return executeLuaFunction(L, name, argument, requireReturn);
}

bool LuaEngine::executeLuaFunction(lua_State *L_p, const char *name, const QVariant &argument, bool requireReturn)
{
    int returnCount = (requireReturn) ? 1 : 0;
    lua_getglobal(L_p, name);
    pushVariant(L_p, argument);
    return (lua_pcall(L_p, 1, returnCount, 0) == 0) ? true : false;
}

bool LuaEngine::executeLuaFunction(const char *name, const QVariantList &args, bool requireReturn)
{
    return executeLuaFunction(L, name, args, requireReturn);
}

bool LuaEngine::executeLuaFunction(lua_State *L_p, const char *name, const QVariantList &args, bool requireReturn)
{
    int returnCount = (requireReturn) ? 1 : 0;
    lua_getglobal(L_p, name);
    for (const QVariant &argument : args) {
        pushVariant(L_p, argument);
    }
    return (lua_pcall(L_p, args.count(), returnCount, 0) == 0) ? true : false;
}

void LuaEngine::pushFunction(const char *name, lua_CFunction function)
{
    pushFunction(L, name, function);
}

void LuaEngine::pushFunction(lua_State *L_p, const char *name, lua_CFunction function)
{
    lua_pushcfunction(L_p, function);
    lua_setglobal(L_p, name);
}

void LuaEngine::pushPointer(const char *name, void *pointer)
{
    pushPointer(L, name, pointer);
}

void LuaEngine::pushPointer(lua_State *L_p, const char *name, void *pointer)
{
    pushPointer(L_p, pointer);
    lua_setglobal(L_p, name);
}

void LuaEngine::pushPointer(void *pointer)
{
    pushPointer(L, pointer);
}

void LuaEngine::pushPointer(lua_State *L_p, void *pointer)
{
    lua_pushlightuserdata(L_p, pointer);
}

void LuaEngine::pushVariant(const char *name, const QVariant &variant)
{
    pushVariant(L, name, variant);
}

void LuaEngine::pushVariant(lua_State *L_p, const char *name, const QVariant &variant)
{
    pushVariant(L_p, variant);
    lua_setglobal(L_p, name);
}

void LuaEngine::pushVariant(const QVariant &variant)
{
    pushVariant(L, variant);
}

void LuaEngine::pushVariant(lua_State *L_p, const QVariant &variant)
{
    if (variant.type() == QVariant::Bool) {
        lua_pushboolean(L_p, variant.toBool());
    }
    else if (variant.type() == QVariant::Int) {
        lua_pushinteger(L_p, variant.toInt());
    }
    else if (variant.type() == QVariant::Double) {
        lua_pushnumber(L_p, variant.toDouble());
    }
    else if (variant.type() == QVariant::String) {
        lua_pushstring(L_p, variant.toString().toUtf8().data());
    }
    else if ((QMetaType::Type)variant.type() == QMetaType::Void || (QMetaType::Type)variant.type() == QMetaType::VoidStar) {
        lua_pushlightuserdata(L_p, variant.value<void*>());
    }
    else {
        lua_pushnil(L_p);
    }
}

QVariant LuaEngine::getVariant(const char *name)
{
    lua_getglobal(L, name);
    return returnVariant();
}

QVariant LuaEngine::getVariant(lua_State *L_p, const char *name)
{
    lua_getglobal(L_p, name);
    return returnVariant(L_p);
}

QVariant LuaEngine::getVariant(int index)
{
    return getVariant(L, index);
}

QVariant LuaEngine::getVariant(lua_State *L_p, int index)
{
    if (lua_isboolean(L_p, index)) {
        return QVariant::fromValue((bool)lua_toboolean(L_p, index));
    }
    else if (lua_isinteger(L_p, index)) {
        return QVariant::fromValue(lua_tointeger(L_p, index));
    }
    else if (lua_isnumber(L_p, index)) {
        return QVariant::fromValue(lua_tonumber(L_p, index));
    }
    else if (lua_isstring(L_p, index)) {
        return QVariant::fromValue(QString(lua_tostring(L_p, index)));
    }
    else if (lua_isuserdata(L_p, index)) {
        return QVariant::fromValue(lua_touserdata(L_p, index));
    }
    else if (lua_isnil(L_p, index)) {
        return QVariant();
    }
    QTextStream(stderr) << "Warning: Didn't catch lua_isnil before empty QVariant got returned" << endl;
    return QVariant();
}

void* LuaEngine::returnPointer()
{
    return returnPointer(L);
}

void* LuaEngine::returnPointer(lua_State *L_p)
{
    return getPointer(L_p, -1);
}

void* LuaEngine::getPointer(const char *name)
{
    lua_getglobal(L, name);
    return returnPointer();
}

void* LuaEngine::getPointer(lua_State *L_p, const char *name)
{
    lua_getglobal(L_p, name);
    return returnPointer(L_p);
}

void* LuaEngine::getPointer(int index)
{
    return getPointer(L, index);
}

void* LuaEngine::getPointer(lua_State *L_p, int index)
{
    return lua_touserdata(L_p, index);
}

QVariant LuaEngine::returnVariant()
{
    return returnVariant(L);
}

QVariant LuaEngine::returnVariant(lua_State *L_p)
{
    return getVariant(L_p, -1);
}

QVariantList LuaEngine::getArguments(lua_State *L_p)
{
    QVariantList arguments;
    int argumentCount = getArgumentCount(L_p);
    for (int i = 1; i < (argumentCount + 1); i++) {
        arguments << getVariant(L_p, i);
    }
    return arguments;
}

int LuaEngine::getArgumentCount(lua_State *L_p)
{
    return lua_gettop(L_p);
}
